#ifndef _STIM_GEN_GEN_SURFACE_CODE_H
#define _STIM_GEN_GEN_SURFACE_CODE_H

#include "stim/circuit/circuit.h"
#include "stim/gen/circuit_gen_params.h"

namespace stim {
GeneratedCircuit generate_surface_code_circuit(const CircuitGenParameters &params);
}

#endif
